@@ -26,6 +26,9 @@
   <mbp@samba.org>, Oct 2000.
   */
 #include "rsync.h"
+#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
+#include <iconv.h>
+#endif
 
 extern int verbose;
 extern int dry_run;
@@ -36,12 +39,18 @@ extern int local_server;
 extern int quiet;
 extern int module_id;
 extern int msg_fd_out;
+extern int allow_8bit_chars;
 extern int protocol_version;
 extern int preserve_times;
+extern int log_format_has_i;
 extern int log_format_has_o_or_i;
 extern int daemon_log_format_has_o_or_i;
+extern mode_t orig_umask;
 extern char *auth_user;
 extern char *log_format;
+#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
+extern iconv_t ic_chck;
+#endif
 
 static int log_initialised;
 static int logfile_was_closed;
@@ -65,8 +74,11 @@ struct {
 	{ RERR_STREAMIO   , "error in rsync protocol data stream" },
 	{ RERR_MESSAGEIO  , "errors with program diagnostics" },
 	{ RERR_IPC        , "error in IPC code" },
-	{ RERR_SIGNAL     , "received SIGUSR1 or SIGINT" },
-	{ RERR_WAITCHILD  , "some error returned by waitpid()" },
+	{ RERR_CRASHED    , "sibling process crashed" },
+	{ RERR_TERMINATED , "sibling process terminated abnormally" },
+	{ RERR_SIGNAL1    , "received SIGUSR1" },
+	{ RERR_SIGNAL     , "received SIGINT, SIGTERM, or SIGHUP" },
+	{ RERR_WAITCHILD  , "waitpid() failed" },
 	{ RERR_MALLOC     , "error allocating core memory buffers" },
 	{ RERR_PARTIAL    , "some files could not be transferred" },
 	{ RERR_VANISHED   , "some files vanished before they could be transferred" },
@@ -80,7 +92,6 @@ struct {
 };
 
 
-
 /*
  * Map from rsync error code to name, or return NULL.
  */
@@ -94,7 +105,6 @@ static char const *rerr_name(int code)
 	return NULL;
 }
 
-
 static void logit(int priority, char *buf)
 {
 	if (logfile_was_closed)
@@ -134,8 +144,7 @@ static void syslog_init()
 
 static void logfile_open(void)
 {
-	extern int orig_umask;
-	int old_umask = umask(022 | orig_umask);
+	mode_t old_umask = umask(022 | orig_umask);
 	logfile = fopen(logfname, "a");
 	umask(old_umask);
 	if (!logfile) {
@@ -187,20 +196,41 @@ void logfile_reopen(void)
 	}
 }
 
+static void filtered_fwrite(FILE *f, const char *buf, int len, int use_isprint)
+{
+	const char *s, *end = buf + len;
+	for (s = buf; s < end; s++) {
+		if ((s < end - 4
+		  && *s == '\\' && s[1] == '#'
+		  && isdigit(*(uchar*)(s+2))
+		  && isdigit(*(uchar*)(s+3))
+		  && isdigit(*(uchar*)(s+4)))
+		 || (*s != '\t'
+		  && ((use_isprint && !isprint(*(uchar*)s))
+		   || *(uchar*)s < ' '))) {
+			if (s != buf && fwrite(buf, s - buf, 1, f) != 1)
+				exit_cleanup(RERR_MESSAGEIO);
+			fprintf(f, "\\#%03o", *(uchar*)s);
+			buf = s + 1;
+		}
+	}
+	if (buf != end && fwrite(buf, end - buf, 1, f) != 1)
+		exit_cleanup(RERR_MESSAGEIO);
+}
+
 /* this is the underlying (unformatted) rsync debugging function. Call
- * it with FINFO, FERROR or FLOG */
+ * it with FINFO, FERROR or FLOG.  Note: recursion can happen with
+ * certain fatal conditions. */
 void rwrite(enum logcode code, char *buf, int len)
 {
+	int trailing_CR_or_NL;
 	FILE *f = NULL;
-	/* recursion can happen with certain fatal conditions */
-
-	if (quiet && code == FINFO)
-		return;
 
 	if (len < 0)
 		exit_cleanup(RERR_MESSAGEIO);
 
-	buf[len] = 0;
+	if (quiet && code == FINFO)
+		return;
 
 	if (am_server && msg_fd_out >= 0) {
 		/* Pass the message to our sibling. */
@@ -208,6 +238,9 @@ void rwrite(enum logcode code, char *buf, int len)
 		return;
 	}
 
+	if (code == FSOCKERR) /* This gets simplified for a non-sibling. */
+		code = FERROR;
+
 	if (code == FCLIENT)
 		code = FINFO;
 	else if (am_daemon) {
@@ -239,31 +272,67 @@ void rwrite(enum logcode code, char *buf, int len)
 		}
 	}
 
-	if (code == FERROR) {
+	switch (code) {
+	case FERROR:
 		log_got_error = 1;
 		f = stderr;
+		goto pre_scan;
+	case FINFO:
+		f = am_server ? stderr : stdout;
+	pre_scan:
+		while (len > 1 && *buf == '\n') {
+			fputc(*buf, f);
+			buf++;
+			len--;
+		}
+		break;
+	case FNAME:
+		f = am_server ? stderr : stdout;
+		break;
+	default:
+		exit_cleanup(RERR_MESSAGEIO);
 	}
 
-	if (code == FINFO)
-		f = am_server ? stderr : stdout;
+	trailing_CR_or_NL = len && (buf[len-1] == '\n' || buf[len-1] == '\r')
+			  ? buf[--len] : 0;
 
-	if (!f)
-		exit_cleanup(RERR_MESSAGEIO);
+#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
+	if (ic_chck != (iconv_t)-1) {
+		char convbuf[1024];
+		char *in_buf = buf, *out_buf = convbuf;
+		size_t in_cnt = len, out_cnt = sizeof convbuf - 1;
 
-	if (fwrite(buf, len, 1, f) != 1)
-		exit_cleanup(RERR_MESSAGEIO);
+		iconv(ic_chck, NULL, 0, NULL, 0);
+		while (iconv(ic_chck, &in_buf,&in_cnt,
+				 &out_buf,&out_cnt) == (size_t)-1) {
+			if (out_buf != convbuf) {
+				filtered_fwrite(f, convbuf, out_buf - convbuf, 0);
+				out_buf = convbuf;
+				out_cnt = sizeof convbuf - 1;
+			}
+			if (errno == E2BIG)
+				continue;
+			fprintf(f, "\\#%03o", *(uchar*)in_buf++);
+			in_cnt--;
+		}
+		if (out_buf != convbuf)
+			filtered_fwrite(f, convbuf, out_buf - convbuf, 0);
+	} else
+#endif
+		filtered_fwrite(f, buf, len, !allow_8bit_chars);
 
-	if (buf[len-1] == '\r' || buf[len-1] == '\n')
+	if (trailing_CR_or_NL) {
+		fputc(trailing_CR_or_NL, f);
 		fflush(f);
+	}
 }
 
-
 /* This is the rsync debugging function. Call it with FINFO, FERROR or
  * FLOG. */
 void rprintf(enum logcode code, const char *format, ...)
 {
 	va_list ap;
-	char buf[MAXPATHLEN+512];
+	char buf[BIGPATHBUFLEN];
 	size_t len;
 
 	va_start(ap, format);
@@ -274,7 +343,7 @@ void rprintf(enum logcode code, const char *format, ...)
 	 * truncate the resulting string.  (Note that configure ensures
 	 * that we have a vsnprintf() that doesn't ever return -1.) */
 	if (len > sizeof buf - 1) {
-		const char ellipsis[] = "[...]";
+		static const char ellipsis[] = "[...]";
 
 		/* Reset length, and zero-terminate the end of our buffer */
 		len = sizeof buf - 1;
@@ -290,7 +359,7 @@ void rprintf(enum logcode code, const char *format, ...)
 		 * If the input format string has a trailing newline,
 		 * we copy it into that extra null; if it doesn't, well,
 		 * all we lose is one byte.  */
-		strncpy(buf+len-sizeof ellipsis, ellipsis, sizeof ellipsis);
+		memcpy(buf+len-sizeof ellipsis, ellipsis, sizeof ellipsis);
 		if (format[strlen(format)-1] == '\n') {
 			buf[len-1] = '\n';
 		}
@@ -299,7 +368,6 @@ void rprintf(enum logcode code, const char *format, ...)
 	rwrite(code, buf, len);
 }
 
-
 /* This is like rprintf, but it also tries to print some
  * representation of the error code.  Normally errcode = errno.
  *
@@ -311,7 +379,7 @@ void rprintf(enum logcode code, const char *format, ...)
 void rsyserr(enum logcode code, int errcode, const char *format, ...)
 {
 	va_list ap;
-	char buf[MAXPATHLEN+512];
+	char buf[BIGPATHBUFLEN];
 	size_t len;
 
 	strcpy(buf, RSYNC_NAME ": ");
@@ -331,8 +399,6 @@ void rsyserr(enum logcode code, int errcode, const char *format, ...)
 	rwrite(code, buf, len);
 }
 
-
-
 void rflush(enum logcode code)
 {
 	FILE *f = NULL;
@@ -360,8 +426,6 @@ void rflush(enum logcode code)
 	fflush(f);
 }
 
-
-
 /* a generic logging routine for send/recv, with parameter
  * substitiution */
 static void log_formatted(enum logcode code, char *format, char *op,
@@ -399,51 +463,83 @@ static void log_formatted(enum logcode code, char *format, char *op,
 		n = NULL;
 
 		switch (*p) {
-		case 'h': if (am_daemon) n = client_name(0); break;
-		case 'a': if (am_daemon) n = client_addr(0); break;
+		case 'h':
+			if (am_daemon)
+				n = client_name(0);
+			break;
+		case 'a':
+			if (am_daemon)
+				n = client_addr(0);
+			break;
 		case 'l':
 			strlcat(fmt, ".0f", sizeof fmt);
 			snprintf(buf2, sizeof buf2, fmt,
 				 (double)file->length);
 			n = buf2;
 			break;
+		case 'U':
+			strlcat(fmt, "ld", sizeof fmt);
+			snprintf(buf2, sizeof buf2, fmt,
+				 (long)file->uid);
+			n = buf2;
+			break;
+		case 'G':
+			if (file->gid == GID_NONE)
+				n = "DEFAULT";
+			else {
+				strlcat(fmt, "ld", sizeof fmt);
+				snprintf(buf2, sizeof buf2, fmt,
+					 (long)file->gid);
+				n = buf2;
+			}
+			break;
 		case 'p':
 			strlcat(fmt, "ld", sizeof fmt);
 			snprintf(buf2, sizeof buf2, fmt,
 				 (long)getpid());
 			n = buf2;
 			break;
-		case 'o': n = op; break;
+		case 'M':
+			n = timestring(file->modtime);
+			{
+				char *cp = n;
+				while ((cp = strchr(cp, ' ')) != NULL)
+					*cp = '-';
+			}
+			break;
+		case 'B':
+			n = buf2 + MAXPATHLEN - PERMSTRING_SIZE;
+			permstring(n - 1, file->mode); /* skip the type char */
+			break;
+		case 'o':
+			n = op;
+			break;
 		case 'f':
-			n = safe_fname(f_name(file));
+			n = f_name(file, NULL);
 			if (am_sender && file->dir.root) {
 				pathjoin(buf2, sizeof buf2,
 					 file->dir.root, n);
-				/* The buffer from safe_fname() has more
-				 * room than MAXPATHLEN, so this is safe. */
+				clean_fname(buf2, 0);
 				if (fmt[1])
-					strcpy(n, buf2);
+					strlcpy(n, buf2, MAXPATHLEN);
 				else
 					n = buf2;
-			}
-			clean_fname(n, 0);
+			} else
+				clean_fname(n, 0);
 			if (*n == '/')
 				n++;
 			break;
 		case 'n':
-			n = safe_fname(f_name(file));
-			if (S_ISDIR(file->mode)) {
-				/* The buffer from safe_fname() has more
-				 * room than MAXPATHLEN, so this is safe. */
-				strcat(n, "/");
-			}
+			n = f_name(file, NULL);
+			if (S_ISDIR(file->mode))
+				strlcat(n, "/", MAXPATHLEN);
 			break;
 		case 'L':
 			if (hlink && *hlink) {
-				n = safe_fname(hlink);
+				n = hlink;
 				strcpy(buf2, " => ");
 			} else if (S_ISLNK(file->mode) && file->u.link) {
-				n = safe_fname(file->u.link);
+				n = file->u.link;
 				strcpy(buf2, " -> ");
 			} else {
 				n = "";
@@ -455,10 +551,18 @@ static void log_formatted(enum logcode code, char *format, char *op,
 			snprintf(buf2 + 4, sizeof buf2 - 4, fmt, n);
 			n = buf2;
 			break;
-		case 'm': n = lp_name(module_id); break;
-		case 't': n = timestring(time(NULL)); break;
-		case 'P': n = lp_path(module_id); break;
-		case 'u': n = auth_user; break;
+		case 'm':
+			n = lp_name(module_id);
+			break;
+		case 't':
+			n = timestring(time(NULL));
+			break;
+		case 'P':
+			n = lp_path(module_id);
+			break;
+		case 'u':
+			n = auth_user;
+			break;
 		case 'b':
 			if (am_sender) {
 				b = stats.total_written -
@@ -494,17 +598,17 @@ static void log_formatted(enum logcode code, char *format, char *op,
 			     : !(iflags & ITEM_TRANSFER) ? '.'
 			     : !local_server && *op == 's' ? '<' : '>';
 			n[1] = S_ISDIR(file->mode) ? 'd'
+			     : IS_SPECIAL(file->mode) ? 'S'
 			     : IS_DEVICE(file->mode) ? 'D'
 			     : S_ISLNK(file->mode) ? 'L' : 'f';
 			n[2] = !(iflags & ITEM_REPORT_CHECKSUM) ? '.' : 'c';
 			n[3] = !(iflags & ITEM_REPORT_SIZE) ? '.' : 's';
 			n[4] = !(iflags & ITEM_REPORT_TIME) ? '.'
-			     : !preserve_times || IS_DEVICE(file->mode)
-					       || S_ISLNK(file->mode) ? 'T' : 't';
+			     : !preserve_times || S_ISLNK(file->mode) ? 'T' : 't';
 			n[5] = !(iflags & ITEM_REPORT_PERMS) ? '.' : 'p';
 			n[6] = !(iflags & ITEM_REPORT_OWNER) ? '.' : 'o';
 			n[7] = !(iflags & ITEM_REPORT_GROUP) ? '.' : 'g';
-			n[8] = !(iflags & ITEM_REPORT_XATTRS) ? '.' : 'a';
+			n[8] = '.';
 			n[9] = '\0';
 
 			if (iflags & (ITEM_IS_NEW|ITEM_MISSING_DATA)) {
@@ -512,7 +616,8 @@ static void log_formatted(enum logcode code, char *format, char *op,
 				int i;
 				for (i = 2; n[i]; i++)
 					n[i] = ch;
-			} else if (!(iflags & (ITEM_TRANSFER|ITEM_LOCAL_CHANGE))) {
+			} else if (n[0] == '.' || n[0] == 'h'
+			        || (n[0] == 'c' && n[1] == 'f')) {
 				int i;
 				for (i = 2; n[i]; i++) {
 					if (n[i] != '.')
@@ -594,7 +699,7 @@ void log_item(struct file_struct *file, struct stats *initial_stats,
 		log_formatted(FLOG, lp_log_format(module_id), s_or_r,
 			      file, initial_stats, iflags, hlink);
 	} else if (log_format && !am_server) {
-		log_formatted(FINFO, log_format, s_or_r,
+		log_formatted(FNAME, log_format, s_or_r,
 			      file, initial_stats, iflags, hlink);
 	}
 }
@@ -602,12 +707,15 @@ void log_item(struct file_struct *file, struct stats *initial_stats,
 void maybe_log_item(struct file_struct *file, int iflags, int itemizing,
 		    char *buf)
 {
-	int see_item = itemizing && (iflags || verbose > 1);
+	int significant_flags = iflags & SIGNIFICANT_ITEM_FLAGS;
+	int see_item = itemizing && (significant_flags || *buf
+		|| log_format_has_i > 1 || (verbose > 1 && log_format_has_i));
+	int local_change = iflags & ITEM_LOCAL_CHANGE && significant_flags;
 	if (am_server) {
 		if (am_daemon && !dry_run && see_item)
 			log_item(file, &stats, iflags, buf);
-	} else if (see_item || iflags & ITEM_LOCAL_CHANGE || *buf
-	    || (S_ISDIR(file->mode) && iflags & SIGNIFICANT_ITEM_FLAGS))
+	} else if (see_item || local_change || *buf
+	    || (S_ISDIR(file->mode) && significant_flags))
 		log_item(file, &stats, iflags, buf);
 }
 
@@ -639,7 +747,6 @@ void log_delete(char *fname, int mode)
 	log_formatted(FLOG, fmt, "del.", &file, &stats, ITEM_DELETED, NULL);
 }
 
-
 /*
  * Called when the transfer is interrupted for some reason.
  *
@@ -662,11 +769,11 @@ void log_exit(int code, const char *file, int line)
 
 		/* VANISHED is not an error, only a warning */
 		if (code == RERR_VANISHED) {
-			rprintf(FINFO, "rsync warning: %s (code %d) at %s(%d)\n", 
-				name, code, file, line);
+			rprintf(FINFO, "rsync warning: %s (code %d) at %s(%d) [%s]\n", 
+				name, code, file, line, who_am_i());
 		} else {
-			rprintf(FERROR, "rsync error: %s (code %d) at %s(%d)\n",
-				name, code, file, line);
+			rprintf(FERROR, "rsync error: %s (code %d) at %s(%d) [%s]\n",
+				name, code, file, line, who_am_i());
 		}
 	}
 }
