@@ -85,7 +85,7 @@
 # they're explicitly given on the command line.
 
 # Also, we can't count on 'cp -a' or 'mkdir -p', although they're
-# pretty handy.
+# pretty handy (see function makepath for the latter).
 
 # I think some of the GNU documentation suggests that we shouldn't
 # rely on shell functions.  However, the Bash manual seems to say that
@@ -100,6 +100,8 @@
 # You cannot do "export VAR=VALUE" all on one line; the export must be
 # separate from the assignment.  (SCO SysV)
 
+# Don't rely on grep -q, as that doesn't work everywhere -- just redirect
+# stdout to /dev/null to keep it quiet.
 
 
 # STILL TO DO:
@@ -124,14 +126,13 @@ set -e
 . "./shconfig"
 
 RUNSHFLAGS='-e'
+export RUNSHFLAGS
 
 # for Solaris
-PATH="/usr/xpg4/bin/:$PATH"
+[ -d /usr/xpg4/bin ] && PATH="/usr/xpg4/bin/:$PATH"
 
-if [ -n "$loglevel" ] && [ "$loglevel" -gt 8 ]
-then
-    if set -x
-    then
+if [ "x$loglevel" != x ] && [ "$loglevel" -gt 8 ]; then
+    if set -x; then
 	# If it doesn't work the first time, don't keep trying.
 	RUNSHFLAGS="$RUNSHFLAGS -x"
     fi
@@ -142,33 +143,45 @@ echo "$0 running in `pwd`"
 echo "    rsync_bin=$rsync_bin"
 echo "    srcdir=$srcdir"
 
-testuser=`whoami || echo UNKNOWN`
+if [ -f /usr/bin/whoami ]; then
+    testuser=`/usr/bin/whoami`
+elif [ -f /usr/ucb/whoami ]; then
+    testuser=`/usr/ucb/whoami`
+elif [ -f /bin/whoami ]; then
+    testuser=`/bin/whoami`
+else
+    testuser=`id -un 2>/dev/null || echo ${LOGNAME:-${USERNAME:-${USER:-'UNKNOWN'}}}`
+fi
 
 echo "    testuser=$testuser"
 echo "    os=`uname -a`"
 
 # It must be "yes", not just nonnull
-if test "x$preserve_scratch" = xyes
-then
+if [ "x$preserve_scratch" = xyes ]; then
     echo "    preserve_scratch=yes"
 else
     echo "    preserve_scratch=no"
 fi    
 
+# We'll use setfacl if it's around and it supports the -k option.
+if setfacl --help 2>/dev/null | grep ' -k,' >/dev/null; then
+    setfacl=setfacl
+else
+    setfacl=true
+fi
 
-if test ! -f $rsync_bin
-then
+if [ ! -f "$rsync_bin" ]; then
     echo "rsync_bin $rsync_bin is not a file" >&2
     exit 2
 fi
 
-if test ! -d $srcdir
-then
+if [ ! -d "$srcdir" ]; then
     echo "srcdir $srcdir is not a directory" >&2
     exit 2
 fi
 
 RSYNC="$rsync_bin"
+#RSYNC="valgrind --tool=addrcheck $rsync_bin"
 
 export rsync_bin RSYNC
 
@@ -190,6 +203,9 @@ export scratchdir suitedir
 prep_scratch() {
     [ -d "$scratchdir" ] && rm -rf "$scratchdir"
     mkdir "$scratchdir"
+    # Get rid of default ACLs and dir-setgid to avoid confusing some tests.
+    $setfacl -k "$scratchdir"
+    chmod g-s "$scratchdir"
     return 0
 }
 
@@ -198,14 +214,13 @@ maybe_discard_scratch() {
     return 0
 }
 
-if [ "x$whichtests" = x ]
-then
+if [ "x$whichtests" = x ]; then
     whichtests="*.test"
 fi
 
 for testscript in $suitedir/$whichtests
 do
-    testbase=`echo $testscript | sed 's!.*/!!' | sed -e 's/.test\$//'`
+    testbase=`echo $testscript | sed -e 's!.*/!!' -e 's/.test\$//'`
     scratchdir="$scratchbase.$testbase"
 
     prep_scratch
@@ -220,6 +235,11 @@ do
 	echo "----- $testbase log follows"
 	cat "$scratchdir/test.log"
 	echo "----- $testbase log ends"
+	if [ -f "$scratchdir/rsyncd.log" ]; then
+	    echo "----- $testbase rsyncd.log follows"
+	    cat "$scratchdir/rsyncd.log"
+	    echo "----- $testbase rsyncd.log ends"
+	fi
     fi
 
     case $result in
@@ -245,8 +265,7 @@ do
     *)
 	echo "FAIL    $testbase"
 	failed=`expr $failed + 1`
-	if [ "x$nopersist" = "xyes" ]
-	then
+	if [ "x$nopersist" = xyes ]; then
 	    exit 1
 	fi
     esac
